# -*- coding: utf-8 -*-
from Products.CMFPlone import PloneMessageFactory as _
from Products.statusmessages.interfaces import IStatusMessage
from collective.easyform.interfaces import ISaveData
from collective.easyform.interfaces import IFieldExtender
from collective.easyform.api import get_schema
from collective.easyform.api import getFieldsInOrder
from collective.easyform.api import get_expression
from collective.easyform.api import cleanup

from plone import api
from plone.memoize.view import memoize
from Products.Five.browser import BrowserView

from six import StringIO
import csv
from DateTime import DateTime

from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IFromUnicode
from zope.schema.interfaces import ISet
from zope.schema.interfaces import IBool

import random
import string

from collective.easyform.browser.view import EasyFormForm as easyformview
from collective.easyform.browser.actions import SavedDataForm
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from plone.z3cform import layout
from z3c.form import button
from collective.easyform.interfaces import ISavedDataFormWrapper

from z3c.form import field
from plone.z3cform.crud import crud
from plone.batching import Batch
import sys
import ast

ATTENDEE_SAVE_DATA_ADAPTER_TITLE = 'Attendee Emails – DO NOT REMOVE'


class ViewAttendees(BrowserView):

    @memoize
    def attendee_emails(self):
        """return a list of the attendee email addresses"""
        all_email_addresses = []
        # read all registration forms in this conference
        brains = api.content.find(context=self.context, portal_type='EasyForm')
        for brain in brains:
            form = brain.getObject()
            # get all the form's save data adapters
            all_save_data_adapters = form.restrictedTraverse("@@get_save_data_adapters")
            results = all_save_data_adapters()
            for saver in results:
                # must have the ID 'attendee_emails___do_not_remove'
                if saver.title == ATTENDEE_SAVE_DATA_ADAPTER_TITLE:
                    save_data_adapter = ISaveData(saver)
                    data = save_data_adapter.getSavedFormInput()
                    # relying on the email address field to have the ID 'replyto'
                    if len(data) > 0 and 'replyto' in data[0].keys():
                        email_addresses = [item['replyto'] for item in data if item['approved'] == "Yes"]
                        all_email_addresses.extend(email_addresses)
        unique_email_addresses_set = set(all_email_addresses)
        return list(unique_email_addresses_set)


    @memoize
    def attendee_list(self):
        """return a list of the attendees"""
        all_email_addresses = []
        all_attendees = []
        # read all registration forms in this conference
        brains = api.content.find(context=self.context, portal_type='EasyForm')
        for brain in brains:
            form = brain.getObject()
            # get all the form's save data adapters
            all_save_data_adapters = form.restrictedTraverse("@@get_save_data_adapters")
            results = all_save_data_adapters()
            for saver in results:
                # must have the ID 'attendee_emails___do_not_remove'
                if saver.title == ATTENDEE_SAVE_DATA_ADAPTER_TITLE:
                    save_data_adapter = ISaveData(saver)
                    data = save_data_adapter.getSavedFormInput()
                    # relying on the email address field to have the ID 'replyto'
                    if len(data) > 0 and 'replyto' in data[0].keys():
                        # create string with Full Name + Email Address
                        keys = ['first_name', 'middle_name', 'last_name']
                        for item in data:
                            full_name = [item[key] for key in keys if item[key]]
                            full_name = ' '.join(filter(None, full_name))
                            attendee_details = (str(full_name), item['replyto'], item['approved'])                            
                            all_attendees.append(attendee_details)
        unique_attendees_set = set(all_attendees)
        new_keys = ('name', 'email', 'approved')
        unique_attendees_list = [dict(zip(new_keys, x)) for x in unique_attendees_set]
        return unique_attendees_list


    @memoize
    def bcc_email_link(self):
        """return a mailto URL"""
        email_addresses = self.attendee_emails()
        email_addresses_string = ";".join([email for email in email_addresses])
        return f"mailto:?bcc={email_addresses_string}"

    @memoize
    def forms_with_attendee_save_data_adapters(self):
        """return only the attendee save data adapters in this conference"""
        forms = []
        brains = api.content.find(context=self.context, portal_type='EasyForm')
        for brain in brains:
            form = brain.getObject()
            # get all the form's save data adapters
            all_save_data_adapters = form.restrictedTraverse("@@get_save_data_adapters")
            results = all_save_data_adapters()
            for saver in results:
                # must have the ID 'attendee_emails___do_not_remove'
                if saver.title == ATTENDEE_SAVE_DATA_ADAPTER_TITLE:
                    forms.append(form)
        return forms

    @memoize
    def attendee_group_name(self):
        conference = self.context
        return f"{conference.id}-attendees"


class UpdateAttendeesGroupView(BrowserView):

    def __call__(self, *args, **kwargs):
        # get the email addresses of attendees
        attendees_view = api.content.get_view('view-attendees', self.context, self.request)
        attendee_emails = attendees_view.attendee_emails()
        attendee_emails_set = set(attendee_emails)
        conference = self.context
        attendees_group_name = f"{conference.id}-attendees"
        # get the users
        attendees_users = []
        for attendee in attendee_emails:
            user = api.user.get(username=attendee)
            if user:
                attendees_users.append(user)
        # look up who is in this conference's attendees group
        group = api.group.get(groupname=attendees_group_name)
        if group:
            member_emails = []
            members = api.user.get_users(groupname=attendees_group_name)
            for member in members:
                email = member.getProperty('email')
                if len(email) > 0:
                    member_emails.append(email)
            member_emails_set = set(member_emails)
        # do a set diff
        attendees_not_yet_in_group = attendee_emails_set - member_emails_set
        attendees_to_remove_from_group = member_emails_set - attendee_emails_set
        if attendees_not_yet_in_group or attendees_to_remove_from_group:
            # add needed attendees to group
            for attendee in attendees_not_yet_in_group:
                user = api.user.get(username=attendee)
                if not user:
                    # Generate a random 25-char password
                    chars = string.ascii_letters + string.digits + "*!?@&$^# %"
                    password_base = ''.join(random.choice(chars) for char in range(25))
                    password = password_base + random.choice("*!?@&$^# %")
                    api.user.create(username=attendee, email=attendee, password=password)
                api.group.add_user(groupname=attendees_group_name, username=attendee)
            # remove unneeded attendees from group
            for attendee in attendees_to_remove_from_group:
                api.group.remove_user(groupname=attendees_group_name, username=attendee)
            IStatusMessage(self.request).add(
                _(u'The attendees group has been updated.'))
            self.request.response.redirect(self.request.HTTP_REFERER)
        else:
            IStatusMessage(self.request).add(
                _(u'No change was made to the attendees group.'))
            self.request.response.redirect(self.request.HTTP_REFERER)


class DownloadCSVTemplateView(BrowserView):
    # display download csv button
    index = ViewPageTemplateFile("templates/download_csv_template.pt")
    review_import_template = ViewPageTemplateFile("templates/review_import.pt")

    def __call__(self):
        return self.template()

    @property
    def template(self):
        if self.request.form.get('file_attachment'):
            return self.review_import_template
        else:
            return self.index


class EventInfo(BrowserView):
    @memoize
    def getEventName(self):
        nav_root = api.portal.get_navigation_root(self.context)
        if nav_root.portal_type == 'conference':
            return nav_root.title
        else:
            return None

    @memoize
    def getEventContactName(self):
        nav_root = api.portal.get_navigation_root(self.context)
        if nav_root.portal_type == 'conference':
            return nav_root.contact_name
        else:
            return None

    @memoize
    def getEventContactEmail(self):
        nav_root = api.portal.get_navigation_root(self.context)
        if nav_root.portal_type == 'conference':
            return nav_root.contact_email
        else:
            return None

class GetCSVTemplateView(BrowserView):

    def __call__(self, *args, **kwargs):
        # download csv template for all form fields
        form = self.context
        response = self.request.RESPONSE
        schema = get_schema(form)
        fields = getFieldsInOrder(schema)
        fieldnames = []
        for f in fields:
            fieldnames.append(f[0])

        now = DateTime().ISO().replace(" ", "-").replace(":", "")

        self.request.RESPONSE.setHeader(
            "Content-type",
            "text/comma-separated-values")
        self.request.RESPONSE.setHeader(
            "Content-Disposition",
            'attachment; filename="{0}_{1}.csv"'.format(form.__name__, now),
        )

        csv_file = StringIO()
        writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
        writer.writeheader()
        return csv_file.getvalue()


class ImportFormCSVView(BrowserView):
    # display download csv button
    index = ViewPageTemplateFile("templates/download_csv_template.pt")

    def __call__(self, *args, **kwargs):
        return self.index

    def importCSV(self, *args, **kwargs):
        """Import CSV information """
        form = self.context
        request = self.request
        attachment = request.form['file_attachment']
        data = self.getCSVdata(attachment, form)

    def getCSVdata(self, attachment, form):
        # get form fields
        schema = get_schema(form)
        fields = getFieldsInOrder(schema)
        fieldnames = []
        for f in fields:
            fieldnames.append(f[0])

        # open csv file
        if hasattr(attachment, 'file'):
            data = attachment.file.read()
            data = data.decode('UTF-8')
            file = data.splitlines()
            reader = csv.DictReader(file)
            rows = list(reader)
            return rows


class ImporterEditForm(crud.EditForm):
    template = ViewPageTemplateFile('templates/importer-crud-table.pt')

    def cleanup_values(self, form, item):
        """ Cleanup incoming data to match easyform format """
        schema = get_schema(form)
        for fname in item:
            if fname in schema:
                field = schema.get(fname)
                if ISet.providedBy(field):
                    try:
                        item[fname] = set(cleanup(item[fname]))
                    except ValueError:
                        pass
                if IBool.providedBy(field) and isinstance(item[fname], str):
                    try:
                        setTrue = any(value.lower() == item[fname].lower()
                                      for value in ['yes', 'true'])
                        if setTrue:
                            item[fname] = True
                        else:
                            item[fname] = False
                    except ValueError:
                        pass
        return item

    @property
    def batch(self):
        items = self.context.get_items()
        global exported_csv_data
        exported_csv_data = items
        # batch_size = self.context.batch_size or sys.maxsize
        batch_size = sys.maxsize
        page = int(self.request.get('%spage' % self.prefix, 0))
        return Batch.fromPagenumber(items, batch_size, page + 1)

    @property
    def getData(self):
        # literal_eval does not handle sets
        form = self.context.context
        schema = get_schema(form)
        items = exported_csv_data
        for item in items:
            fid, fields = item
            for fname in fields:
                if fname in schema:
                    field = schema.get(fname)
                    if ISet.providedBy(field):
                        try:
                            fields[fname] = list(fields[fname])
                            fields[fname] = ','.join(str(s)
                                                     for s in fields[fname])
                        except ValueError:
                            pass
                    elif IDate.providedBy(field) and fields[fname]:
                        fields[fname] = fields[fname].strftime("%Y/%m/%d")
                    elif IDatetime.providedBy(field) and fields[fname]:
                        fields[fname] = fields[fname].strftime(
                            "%Y/%m/%d %H:%M:%S")
        return items

    @button.buttonAndHandler(_('Import form Data'),
                             name='edit',
                             condition=lambda form: form.context.update_schema)
    def handle_edit(self, action):
        success = _(u"Successfully updated")
        partly_success = _(u"Some of your changes could not be applied.")
        status = no_changes = _(u"No changes made.")
        for subform in self.subforms:
            # With the ``extractData()`` call, validation will occur,
            # and errors will be stored on the widgets amongst other
            # places.  After this we have to be extra careful not to
            # call (as in ``__call__``) the subform again, since
            # that'll remove the errors again.  With the results that
            # no changes are applied but also no validation error is
            # shown.
            data, errors = subform.extractData()
            if errors:
                if status is no_changes:
                    status = subform.formErrorsMessage
                elif status is success:
                    status = partly_success
                return
            # del data['select']
            # self.context.before_update(subform.content, data)
            # submit data if no errors
            imported_rows = 0
            form = self.context.context
            rows = self.subforms
            total_rows = len(rows)
            for x in rows:
                row = data
                try:
                    csvimportcrud(self.context,
                                  self.request).submitForm(form, row)
                    imported_rows += 1
                except IndexError:
                    continue
            # self.status = status
            IStatusMessage(self.request).add(
                    _(u'Imported {0} out of {1} row(s).').format(imported_rows,
                                                                 total_rows))


class csvimportcrud(SavedDataForm):
    """ Use the SavedData Adapter form to review and validate csv import data
    """

    template = ViewPageTemplateFile("templates/review_import.pt")
    editform_factory = ImporterEditForm

    def importCSV(self, *args, **kwargs):
        """Import CSV information """
        form = self.context
        request = self.request
        attachment = request.form['file_attachment']
        data = self.getCSVdata(attachment, form)
        return data

    def getCSVdata(self, attachment, form):
        # get form fields
        schema = get_schema(form)
        fields = getFieldsInOrder(schema)
        fieldnames = []
        for f in fields:
            fieldnames.append(f[0])

        # open csv file
        if hasattr(attachment, 'file'):
            data = attachment.file.read()
            data = data.decode('UTF-8')
            file = data.splitlines()
            reader = csv.DictReader(file)
            rows = list(reader)
            # # format to match SavedDataForm expects
            formatted_rows = []
            for row in rows:
                unsorted_data = dict(row)
                unsorted_data = self.cleanup_values(form, row)
                formatted_rows.append(unsorted_data)

            return [(count, dict(formatted_rows))
                    for count, formatted_rows in enumerate(formatted_rows, 1)]

    def submitForm(self, form, row):
        """ Process Easyform actions """
        # unsorted_data = dict(row)
        unsorted_data = row
        # Default 'Approved' to 'No' if unset in import
        if 'approved' in unsorted_data:
            if unsorted_data['approved'] not in ['Yes', 'No']:
                unsorted_data['approved'] = 'No'
        errors = easyformview.processActions(self.context, unsorted_data)
        if errors:
            return form.setErrorsMessage(errors)

    def cleanup_values(self, form, data):
        """ Cleanup incoming data to match easyform format """
        schema = get_schema(form)
        for fname in data:
            if fname in schema:
                field = schema.get(fname)
                if ISet.providedBy(field) and data[fname]:
                    try:
                        data[fname] = set(cleanup(data[fname]))
                    except ValueError:
                        pass
                elif IBool.providedBy(field) and isinstance(data[fname], str):
                    try:
                        setTrue = any(value.lower() == data[fname].lower()
                                      for value in ['yes', 'true', 'y'])
                        if setTrue:
                            data[fname] = True
                        else:
                            data[fname] = False
                    except ValueError:
                        pass
                elif IDatetime.providedBy(field) and data[fname]:
                    try:
                        data[fname] = DateTime(data[fname]).asdatetime()
                    except ValueError:
                        pass
                elif IDate.providedBy(field) and data[fname]:
                    try:
                        data[fname] = DateTime(data[fname]).asdatetime().date()
                    except ValueError:
                        pass
        return data

    @property
    def field(self):
        # Define Form display settings
        field_data = {}
        return field_data

    def get_items(self):
        """ Get csv data and format """
        data = []
        if hasattr(self.request, 'file_attachment'):
            if self.request.form.get('file_attachment'):
                data = self.importCSV()
                return data
        elif hasattr(self.request, 'crud-edit.form.buttons.edit'):
            # get data from submitted form
            if self.request.form.get('data_export'):
                data_str = self.request.data_export
                # ast cannot handle sets
                data = ast.literal_eval(data_str)
        else:
            data = []
            url = "{0}/@@import-forms".format(
                self.context.absolute_url()
            )
            IStatusMessage(self.request).add(
                _(u'Error: Missing file attachment.'))
            self.request.response.redirect(url)
        return data

    @property
    def get_schema(self):
        form = self.context
        return get_schema(form)

    @property
    def view_schema(self):
        """ ignore Saved Data ExtraData fields
        """
        pass


class ImportDataFormWrapper(layout.FormWrapper):
    def __call__(self):
        if hasattr(self.request, 'file_attachment'):
            return super(ImportDataFormWrapper, self).__call__()
        elif hasattr(self.request, 'form'):
            return super(ImportDataFormWrapper, self).__call__()
        else:
            url = "{0}/@@import-forms".format(
                self.context.absolute_url()
            )
            IStatusMessage(self.request).add(
                _(u'Error: Missing file attachment.'))
            self.request.response.redirect(url)


PreviewCsvDataView = layout.wrap_form(
    # csvimportcrud, __wrapper_class=ImportDataFormWrapper
    csvimportcrud
)
